from kucoin.base_request.base_request import KucoinBaseRestApi


class MarginData(KucoinBaseRestApi):

    def get_mark_price(self, symbol):
        """
        https://docs.kucoin.com/#margin-info
        :param symbol: symbol (Mandatory)
        :type: str
        :return:
        {
            "symbol": "USDT-BTC",
            "granularity": 5000,
            "timePoint": 1568701710000,
            "value": 0.00009807
        }
        """
        return self._request('GET', '/api/v1/mark-price/{symbol}/current'.format(symbol=symbol))

    def get_margin_config(self):
        """
        https://docs.kucoin.com/#get-margin-configuration-info
        :return:
        {
            "currencyList": ["BTC","USDT","EOS"],
            "warningDebtRatio": "0.8",
            "liqDebtRatio": "0.9",
            "maxLeverage": "3"
        }
        """
        return self._request('GET', '/api/v1/margin/config')

    def get_margin_account(self):
        """
        https://docs.kucoin.com/#get-margin-account
        :return:
        {
            "accounts": [
              {
                "availableBalance": "990.11",
                "currency": "USDT",
                "holdBalance": "7.22",
                "liability": "66.66",
                "maxBorrowSize": "88.88",
                "totalBalance": "997.33"
              }
            ],
            "debtRatio": "0.33"
        }
        """
        return self._request('GET', '/api/v1/margin/account')

    def create_borrow_order(self, currency, order_type, size, **kwargs):
        """
        https://docs.kucoin.com/#post-borrow-order
        :param currency: Currency to Borrow (Mandatory)
        :type: str
        :param order_type: Type: FOK, IOC (Mandatory)
        :type: str
        :param size: Total size (Mandatory)
        :type: float
        :param kwargs: [Optional] maxRate, term
        :return:
        {
            "orderId": "a2111213",
            "currency": "USDT"
        }
        """
        params = {
            'currency': currency,
            'type': order_type,
            'size': size,
        }
        if kwargs:
            params.update(kwargs)
        return self._request('POST', '/api/v1/margin/borrow', params=params)

    def get_borrow_order(self, orderId):
        """
        https://docs.kucoin.com/#get-borrow-order
        :param orderId: Borrow order ID
        :type: str
        :return:
        {
            "currency": "USDT",
            "filled": 1.009,
            "matchList": [
              {
                "currency": "USDT",
                "dailyIntRate": "0.001",
                "size": "12.9",
                "term": 7,
                "timestamp": "1544657947759",
                "tradeId": "1212331"
              }
            ],
            "orderId": "a2111213",
            "size": "1.009",
            "status": "DONE"
          }
        """
        params = {
            'orderId': orderId
        }

        return self._request('GET', '/api/v1/margin/borrow', params=params)

    def get_repay_record(self, **kwargs):
        """
        https://docs.kucoin.com/#get-repay-record
        :param kwargs: [Optional] currency, currentPage, pageSize
        :return:
        {
            "currentPage": 0,
            "items": [
              {
                "accruedInterest": "0.22121",
                "createdAt": "1544657947759",
                "currency": "USDT",
                "dailyIntRate": "0.0021",
                "liability": "1.32121",
                "maturityTime": "1544657947759",
                "principal": "1.22121",
                "repaidSize": "0",
                "term": 7,
                "tradeId": "1231141"
              }
            ],
            "pageSize": 0,
            "totalNum": 0,
            "totalPage": 0
          }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/margin/borrow/outstanding', params=params)

    def get_repayment_record(self, **kwargs):
        """
        https://docs.kucoin.com/#get-repayment-record
        :param kwargs: [Optional] currency, currentPage, pageSize
        :return:
        {
            "currentPage": 0,
            "items": [
              {
                "currency": "USDT",
                "dailyIntRate": "0.0021",
                "interest": "0.22121",
                "principal": "1.22121",
                "repaidSize": "0",
                "repayTime": "1544657947759",
                "term": 7,
                "tradeId": "1231141"
              }
            ],
            "pageSize": 0,
            "totalNum": 0,
            "totalPage": 0
          }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/margin/borrow/repaid', params=params)

    def click_to_repayment(self, currency, sequence, size):
        """
        https://docs.kucoin.com/#one-click-repayment
        :param currency: currency (Mandatory)
        :type: str
        :param sequence: Repayment strategy. (Mandatory)
        RECENTLY_EXPIRE_FIRST: Time priority, namely to repay the loans of the nearest maturity time first,
        HIGHEST_RATE_FIRST: Rate Priority: Repay the loans of the highest interest rate first.
        :type: str
        :param size: Repayment size (Mandatory)
        :type: float
        :return:
        """
        params = {
            'currency': currency,
            'sequence': sequence,
            'size': size
        }
        return self._request('POST', '/api/v1/margin/repay/all', params=params)

    def repay_single_order(self, currency, tradeId, size):
        """
        https://docs.kucoin.com/#repay-a-single-order
        :param currency: currency (Mandatory)
        :type: str
        :param tradeId: Trade ID (Mandatory)
        :type: str
        :param size: Repayment size (Mandatory)
        :type: float
        :return:
        """
        params = {
            'currency': currency,
            'tradeId': tradeId,
            'size': size
        }
        return self._request('POST', '/api/v1/margin/repay/single', params=params)

    def create_lend_order(self, currency, size, dailyIntRate, term):
        """
        https://docs.kucoin.com/#post-lend-order
        :param currency: Currency to lend (Mandatory)
        :type: str
        :param size: Total size (Mandatory)
        :type: str
        :param dailyIntRate: Daily interest rate. e.g. 0.002 is 0.2% (Mandatory)
        :type: str
        :param term: Term (Unit: Day) (Mandatory)
        :type: int
        :return:
        {
            "orderId": "5da5a4f0f943c040c2f8501e"
        }
        """
        params = {
            'currency': currency,
            'size': size,
            'dailyIntRate': dailyIntRate,
            'term': term
        }
        return self._request('POST', '/api/v1/margin/lend', params=params)

    def cancel_lend_order(self, orderId):
        """
        https://docs.kucoin.com/#cancel-lend-order
        :param orderId: Lend order ID (Mandatory)
        :type: str
        :return:
        """
        return self._request('DELETE', '/api/v1/margin/lend/{orderId}'.format(orderId=orderId))

    def set_auto_lend(self, currency, isEnable, **kwargs):
        """
        https://docs.kucoin.com/#set-auto-lend
        :param currency: currency (Mandatory)
        :type: str
        :param isEnable: Auto-lend enabled or not (Mandatory)
        :type: bool
        :param kwargs: [Required when isEnable is true] retainSize, dailyIntRate, term
        :return:
        """
        params = {
            'currency': currency,
            'isEnable': isEnable
        }
        if kwargs:
            params.update(kwargs)
        return self._request('POST', '/api/v1/margin/toggle-auto-lend', params=params)

    def get_active_order(self, **kwargs):
        """
        https://docs.kucoin.com/#get-active-order
        :param kwargs: [Optional] currency, currentPage, pageSize
        :return:
        {
            "currentPage": 1,
            "pageSize": 1,
            "totalNum": 1,
            "totalPage": 1,
            "items": [{
                "orderId": "5da59f5ef943c033b2b643e4",
                "currency": "BTC",
                "size": "0.51",
                "filledSize": "0",
                "dailyIntRate": "0.0001",
                "term": 7,
                "createdAt": 1571135326913
            }]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/margin/lend/active', params=params)

    def get_lent_history(self, **kwargs):
        """
        https://docs.kucoin.com/#get-lent-history
        :param kwargs: [Optional] currency, currentPage, pageSize
        :return:
        {
            "currentPage": 1,
            "pageSize": 1,
            "totalNum": 1,
            "totalPage": 1,
            "items": [{
                "orderId": "5da59f5bf943c033b2b643da",
                "currency": "BTC",
                "size": "0.51",
                "filledSize": "0.51",
                "dailyIntRate": "0.0001",
                "term": 7,
                "createdAt": 1571135323984,
                "status": "FILLED"
            }]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/margin/lend/done', params=params)

    def get_active_list(self, **kwargs):
        """
        https://docs.kucoin.com/#get-active-lend-order-list
        :param kwargs: [Optional] currency, currentPage, pageSize
        :return:
        {
            "currentPage": 1,
            "pageSize": 1,
            "totalNum": 1,
            "totalPage": 1,
            "items": [{
                "tradeId": "5da6dba0f943c0c81f5d5db5",
                "currency": "BTC",
                "size": "0.51",
                "accruedInterest": "0",
                "repaid": "0.10999968",
                "dailyIntRate": "0.0001",
                "term": 14,
                "maturityTime": 1572425888958
            }]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/margin/lend/trade/unsettled', params=params)

    def get_settled_order(self, **kwargs):
        """
        https://docs.kucoin.com/#get-settled-lend-order-history
        :param kwargs: [Optional] currency, currentPage, pageSize
        :return:
        {
            "currentPage": 1,
            "pageSize": 1,
            "totalNum": 1,
            "totalPage": 1,
            "items": [{
                "tradeId": "5da59fe6f943c033b2b6440b",
                "currency": "BTC",
                "size": "0.51",
                "interest": "0.00004899",
                "repaid": "0.510041641",
                "dailyIntRate": "0.0001",
                "term": 7,
                "settledAt": 1571216254767,
                "note": "The account of the borrowers reached a negative balance, and the system has supplemented the loss via the insurance fund. Deposit funds: 0.51."
            }]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/margin/lend/trade/settled', params=params)

    def get_lend_record(self, currency=None):
        """
        https://docs.kucoin.com/#get-account-lend-record
        :param currency: currency (Optional)
        :type: str
        :return:
        [{
            "currency": "BTC",
            "outstanding": "1.02",
            "filledSize": "0.91000213",
            "accruedInterest": "0.00000213",
            "realizedProfit": "0.000045261",
            "isAutoLend": false
        }]
        """
        params = {}
        if currency:
            params['currency'] = currency
        return self._request('GET', '/api/v1/margin/lend/assets', params=params)

    def get_lending_market(self, currency, term=None):
        """
        https://docs.kucoin.com/#lending-market-data
        :param currency: currency (Mandatory)
        :type: str
        :param term: Term (Unit: Day) (Optional)
        :type: int
        :return:
        [{
            "dailyIntRate": "0.0001",
            "term": 7,
            "size": "1.02"
        }]
        """
        params = {
            'currency': currency
        }
        if term:
            params['term'] = term
        return self._request('GET', '/api/v1/margin/market', params=params)

    def get_margin_data(self, currency):
        """
        https://docs.kucoin.com/#margin-trade-data
        :param currency: currency (Mandatory)
        :type: str
        :return:
        [{
            "tradeId": "5da6dba0f943c0c81f5d5db5",
            "currency": "BTC",
            "size": "0.51",
            "dailyIntRate": "0.0001",
            "term": 14,
            "timestamp": 1571216288958989641
        }]
        """
        params = {
            'currency': currency
        }
        return self._request('GET', '/api/v1/margin/trade/last', params=params)
    
    def get_margin_risk_limit(self, marginModel='cross'):
        """
        https://docs.kucoin.com/#margin-trade-data
        :param marginModel: marginModel
        :type: str
        :return:
        [{
        "currency": "BTC",
        "borrowMaxAmount": "50",
        "buyMaxAmount": "50",
        "precision": 8
        },
        {
        "currency": "SKL",
        "borrowMaxAmount": "50000",
        "buyMaxAmount": "51000",
        "precision": 3
        },
        {
        "currency": "USDT",
        "borrowMaxAmount": "100000",
        "buyMaxAmount": "10000000000",
        "precision": 8
        },
        {
        "currency": "ETH",
        "borrowMaxAmount": "236",
        "buyMaxAmount": "500",
        "precision": 8
        },
        {
        "currency": "LTC",
        "borrowMaxAmount": "100",
        "buyMaxAmount": "40",
        "precision": 8
        }]
        """
        params = {
            'marginModel': marginModel,
        }
        return self._request('GET', '/api/v1/risk/limit/strategy', params=params)




from Exchanges.BaseExchange import BaseExchange

from kucoin.client import User, Trade, Market
from kucoin_futures.client import FuturesUser, FuturesTrade, FuturesMarket


class KuCoinExchange(BaseExchange):
    def __init__(self, credentials, sandbox=False, unifiedInOuts=True):
        self.apiKey = credentials['spot']['key']
        self.secret = credentials['spot']['secret']
        self.passphrase = credentials['spot']['passphrase']
        self.sandbox = sandbox
        self.unifiedInOuts = unifiedInOuts

        if sandbox:
            self.spotUser = User(key=self.apiKey, secret=self.secret, passphrase=self.passphrase, is_sandbox=True)
            self.spotTrade = Trade(key=self.apiKey, secret=self.secret, passphrase=self.passphrase, is_sandbox=True)
            self.spotMarket = Market(key=self.apiKey, secret=self.secret, passphrase=self.passphrase, is_sandbox=True)

            self.futuresUser = FuturesUser(key=self.apiKey, secret=self.secret, passphrase=self.passphrase,
                                           is_sandbox=True)
            self.futuresTrade = FuturesTrade(key=self.apiKey, secret=self.secret, passphrase=self.passphrase,
                                             is_sandbox=True)
            self.futuresMarket = FuturesMarket(key=self.apiKey, secret=self.secret, passphrase=self.passphrase,
                                               is_sandbox=True)
        else:
            self.spotUser = User(key=self.apiKey, secret=self.secret, passphrase=self.passphrase)
            self.spotTrade = Trade(key=self.apiKey, secret=self.secret, passphrase=self.passphrase)
            self.spotMarket = Market(key=self.apiKey, secret=self.secret, passphrase=self.passphrase)

            self.futuresUser = FuturesUser(key=self.apiKey, secret=self.secret, passphrase=self.passphrase)
            self.futuresTrade = FuturesTrade(key=self.apiKey, secret=self.secret, passphrase=self.passphrase)
            self.futuresMarket = FuturesMarket(key=self.apiKey, secret=self.secret, passphrase=self.passphrase)

    def getBalance(self, asset='', futures=False):
        if futures:
            pass

    def symbolAccountTradeHistory(self, symbol, futures=False, fromId=None, limit=None):
        pass

    def testSpotOrder(self, orderData):
        pass

    def makeSpotOrder(self, orderData):
        pass

    def getSymbolOrders(self, symbol, futures=False, orderId=None, startTime=None, endTime=None, limit=None):
        pass

    def getOpenOrders(self, symbol, futures=False):
        pass

    def cancelAllSymbolOpenOrders(self, symbol, futures=False):
        pass

    def cancelOrder(self, symbol, orderId=None, localOrderId=None, futures=False):
        pass

    def getOrder(self, symbol, orderId=None, localOrderId=None, futures=False):
        pass

    def getTradingFees(self):
        pass

    def getSymbolTickerPrice(self, symbol, futures=False):
        pass

    def getSymbolKlines(self, symbol, interval, startTime=None, endTime=None, limit=None, futures=False, blvtnav=False,
                        convertDateTime=False, doClean=False, toCleanDataframe=False):
        pass

    def getExchangeTime(self, futures=False):
        pass

    def getSymbol24hTicker(self, symbol):
        pass

    def testFuturesOrder(self, futuresOrderData):
        pass

    def makeFuturesOrder(self, futuresOrderData):
        pass

    def makeBatchFuturesOrder(self, futuresOrderDatas):
        pass

    def changeInitialLeverage(self, symbol, leverage):
        pass

    def changeMarginType(self, symbol, marginType, params):
        pass

    def changePositionMargin(self, symbol, amount, marginType=None):
        pass

    def getPosition(self):
        pass

    def spotBestBidAsks(self, symbol=None):
        pass

    def getSymbolOrderBook(self, symbol, limit=None, futures=False):
        pass

    def getSymbolRecentTrades(self, symbol, limit=None, futures=False):
        pass

    def getPositionInfo(self, symbol=None):
        pass

    def getSymbolMinTrade(self, symbol, futures=False):
        pass



class Income:

    def __init__(self):
        self.symbol = ""
        self.incomeType = ""
        self.income = 0.0
        self.asset = ""
        self.time = 0
    
    @staticmethod
    def json_parse(json_data):
        result = Income()
        result.symbol = json_data.get_string("symbol")
        result.incomeType = json_data.get_string("incomeType")
        result.income = json_data.get_float("income")
        result.asset = json_data.get_string("asset")
        result.time = json_data.get_int("time")

        return result



class CodeMsg:

    def __init__(self):
        self.code = 0
        self.msg = ""

    @staticmethod
    def json_parse(json_data):
        result = CodeMsg()
        result.code = json_data.get_int("code")
        result.msg = json_data.get_string("msg")

        return result

    def toDict(self):
        outDict = {
            'code': self.code,
            'msg': self.msg
        }

        return outDict



class MarkPriceEvent:

    def __init__(self):
        self.eventType = ""
        self.eventTime = 0
        self.symbol = ""
        self.markPrice = 0.0
        self.fundingRate = 0.0
        self.nextFundingTime = 0

    @staticmethod
    def json_parse(json_data):
        result = MarkPriceEvent()
        result.eventType = json_data.get_string("e")
        result.eventTime = json_data.get_int("E")
        result.symbol = json_data.get_string("s")
        result.markPrice = json_data.get_float("p")
        result.fundingRate = json_data.get_float("r")
        result.nextFundingTime = json_data.get_int("T")
        return result




class WebSocketDefine:
#    Uri = "wss://stream.binancefuture.com/ws"
    Uri = "wss://fstream.binance.com/ws"

class RestApiDefine:
#     Url = "https://testnet.binancefuture.com"
 #   Uri = "wss://stream.binancefuture.com/ws"
    Uri = "wss://fstream.binance.com/ws"

class RestApiDefine:
 #   Url = "https://testnet.binancefuture.com"
    Url = "https://fapi.binance.com"








import json
from datetime import datetime

import pandas as pd

from Utils import DataHelpers


def getSpotOrderAsDict(order: DataHelpers.OrderData):
    if order.timestamp is None:
        raise ValueError('Timestamp must be set')

    params = {'symbol': order.symbol, 'side': order.side, 'type': order.orderType, 'timestamp': order.timestamp}

    if order.timeInForce is not None:
        params['timeInForce'] = order.timeInForce

    if order.quantity is not None:
        params['quantity'] = order.quantity

    if order.quoteOrderQty is not None:
        params['quoteOrderQty'] = order.quoteOrderQty

    if order.price is not None:
        params['price'] = order.price

    if order.newOrderRespType is not None:
        params['newOrderRespType'] = order.newOrderRespType

    if order.stopPrice is not None:
        params['stopPrice'] = order.stopPrice

    if order.icebergQty is not None:
        params['icebergQty'] = order.icebergQty

    if order.newClientOrderId is not None:
        params['newClientOrderId'] = order.newClientOrderId

    if order.recvWindow is not None:
        params['recvWindow'] = order.recvWindow

    return params


def getFuturesOrderAsDict(order: DataHelpers.futuresOrderData, allStr=False):
    params = {'symbol': order.symbol, 'side': order.side, 'ordertype': order.orderType}

    if order.positionSide is not None:
        params['positionSide'] = order.positionSide

    if order.timeInForce is not None:
        params['timeInForce'] = order.timeInForce

    if order.quantity is not None:
        params['quantity'] = order.quantity

    if order.reduceOnly is not None:
        params['reduceOnly'] = order.reduceOnly

    if order.price is not None:
        params['price'] = order.price

    if order.newClientOrderId is not None:
        params['newClientOrderId'] = order.newClientOrderId

    if order.stopPrice is not None:
        params['stopPrice'] = order.stopPrice

    if order.closePosition is not None:
        params['closePosition'] = order.closePosition

    if order.activationPrice is not None:
        params['activationPrice'] = order.activationPrice

    if order.callbackRate is not None:
        params['callbackRate'] = order.callbackRate

    if order.workingType is not None:
        params['workingType'] = order.workingType

    if order.priceProtect is not None:
        params['priceProtect'] = order.priceProtect

    if order.newOrderRespType is not None:
        params['newOrderRespType'] = order.newOrderRespType

    if order.recvWindow is not None:
        params['recvWindow'] = order.recvWindow

    if allStr:
        for key, value in params.items():
            params[key] = str(value)

    return params


def getKlinesDesiredOnlyCols(data, desiredIndexes):
    finalDataArray = []
    for datum in data:
        finalDataArray.append([datum[index] for index in desiredIndexes])
    return finalDataArray


def klinesConvertToPandas(outArray):
    df = pd.DataFrame(outArray,
                      columns=['date', 'open', 'high', 'low', 'close', 'volume', 'closeDate', 'tradesNum'])
    df.set_index('date', inplace=True)
    return df


def klinesConvertDate(data, timeColIdxs):
    for datum in data:
        for idx in timeColIdxs:
            datum[idx] = datetime.fromtimestamp(float(datum[idx]) / 1000)


def extractSymbolInfoFromFilters(symbolFilters, tickerPrice):
    params = {}
    for symbolFilter in symbolFilters:
        if symbolFilter['filterType'] == 'LOT_SIZE':
            params['minQuantity'] = float(symbolFilter['minQty'])
            params['precisionStep'] = float(symbolFilter['stepSize'])
            params['minQuoteQuantity'] = tickerPrice * params['minQuantity']

        if symbolFilter['filterType'] == 'PRICE_FILTER':
            params['stepPrice'] = symbolFilter['tickSize']
    return params


def makeBatchOrderData(futuresOrderDatas):
    batchOrders = []
    for order in futuresOrderDatas:
        orderAsDict = getFuturesOrderAsDict(order, allStr=True)
        orderAsDict['type'] = orderAsDict.pop('ordertype')

        orderJSON = json.dumps(orderAsDict)

        batchOrders.append(orderJSON)
    return batchOrders



import json
from binance_f.impl.utils.timeservice import get_current_timestamp
from binance_f.model import DepthStep


def aggregate_trade_channel(symbol):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@aggTrade")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def mark_price_channel(symbol):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@markPrice")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def continuous_kline_channel(pair, contract_type, interval):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(pair + '_' + contract_type + "@continuousKline_" + interval)
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def kline_channel(symbol, interval):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@kline_" + interval)
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def trade_channel(symbol):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@trade")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def symbol_miniticker_channel(symbol):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@miniTicker")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def all_miniticker_channel():
    channel = dict()
    channel["params"] = list()
    channel["params"].append("!miniTicker@arr")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def symbol_ticker_channel(symbol):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@ticker")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def all_ticker_channel():
    channel = dict()
    channel["params"] = list()
    channel["params"].append("!ticker@arr")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def symbol_bookticker_channel(symbol):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@bookTicker")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def all_bookticker_channel():
    channel = dict()
    channel["params"] = list()
    channel["params"].append("!bookTicker")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def symbol_liquidation_channel(symbol):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@forceOrder")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def all_liquidation_channel(symbol):
    channel = dict()
    channel["params"] = list()
    channel["params"].append("!forceOrder@arr")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def book_depth_channel(symbol, limit, update_time):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@depth" + str(limit) + str(update_time))
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def diff_depth_channel(symbol, update_time):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@depth" + update_time)
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)
    
def user_data_channel(listenKey):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(listenKey)
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def all_mark_price_channel():
    channel = dict()
    channel["params"] = list()
    channel["params"].append("!markPrice@arr")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def blvt_info_channel(symbol):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@compositeIndex")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def blvt_nav_kline_channel(symbol, interval):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@kline_" + interval)
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)

def composite_index_channel(symbol):
    channel = dict()
    channel["params"] = list()
    channel["params"].append(symbol + "@compositeIndex")
    channel["id"] = get_current_timestamp()
    channel["method"] = "SUBSCRIBE"
    return json.dumps(channel)



class Leverage:

    def __init__(self):
        self.leverage = 0.0
        self.maxNotionalValue = 0.0
        self.symbol = ""

    @staticmethod
    def json_parse(json_data):
        result = Leverage()
        result.leverage = json_data.get_float("leverage")
        result.maxNotionalValue = json_data.get_float("maxNotionalValue")
        result.symbol = json_data.get_string("symbol")

        return result

    def toDict(self):
        return {'leverage': self.leverage, 'maxNotionalValue': self.maxNotionalValue, 'symbol': self.symbol}



class MyTrade:

    def __init__(self):
        self.isBuyer = False
        self.commission = 0.0
        self.commissionAsset = ""
        self.counterPartyId = None
        self.id = None
        self.isMaker = False
        self.orderId = None
        self.price = 0.0
        self.qty = 0.0
        self.quoteQty = 0.0
        self.realizedPnl = 0.0
        self.side = ""
        self.symbol = ""
        self.time = 0

    @staticmethod
    def json_parse(json_data):
        result = MyTrade()
        result.isBuyer = json_data.get_boolean("buyer")
        result.commission = json_data.get_float("commission")
        result.commissionAsset = json_data.get_string("commissionAsset")
        result.counterPartyId = json_data.get_int_or_default("counterPartyId", None)
        result.id = json_data.get_int("id")
        result.isMaker = json_data.get_boolean("maker")
        result.orderId = json_data.get_int("orderId")
        result.price = json_data.get_float("price")
        result.qty = json_data.get_float("qty")
        result.quoteQty = json_data.get_float("quoteQty")
        result.realizedPnl = json_data.get_float("realizedPnl")
        result.side = json_data.get_string("side")
        result.symbol = json_data.get_string("symbol")
        result.time = json_data.get_int("time")
        
        return result

    def toDict(self):
        outDict = {'symbol': self.symbol,
                    'id': self.id,
                    'orderId': self.orderId,
                    'orderListId': self.counterPartyId,
                    'price': self.price,
                    'qty': self.qty,
                    'quoteQty': self.quoteQty,
                    'commission': self.commission,
                    'commissionAsset': self.commissionAsset,
                    'time': self.time,
                    'isBuyer': self.isBuyer,
                    'isMaker': self.isMaker,
                    'isBestMatch': None}

        return outDict

# {'symbol': 'BTCUSDT',
#   'id': 12786,
#   'orderId': 15005,
#   'orderListId': -1,
#   'price': '35000.00000000',
#   'qty': '0.00200000',
#   'quoteQty': '70.00000000',
#   'commission': '0.00000000',
#   'commissionAsset': 'BTC',
#   'time': 1646213286609,
#   'isBuyer': True,
#   'isMaker': True,
#   'isBestMatch': True}



class SymbolOrderBook:

    def __init__(self):
        self.symbol = ""
        self.bidPrice = 0.0
        self.bidQty = 0.0
        self.askPrice = 0.0
        self.askQty = 0.0

    @staticmethod
    def json_parse(json_data):
        result = SymbolOrderBook()
        result.symbol = json_data.get_string("symbol")
        result.bidPrice = json_data.get_float("bidPrice")
        result.bidQty = json_data.get_float("bidQty")
        result.askPrice = json_data.get_float("askPrice")
        result.askQty = json_data.get_float("askQty")
        return result


from kucoin.base_request.base_request import KucoinBaseRestApi


class TradeData(KucoinBaseRestApi):

    def create_limit_margin_order(self, symbol, side, size, price, clientOid='', **kwargs):
        """
        https://docs.kucoin.com/#place-a-margin-order
        :param symbol: a valid trading symbol code (Mandatory)
        :type: str
        :param side: place direction buy or sell (Mandatory)
        :type: str
        :param size: amount of base currency to buy or sell (Mandatory)
        :type: str
        :param price: price per base currency (Mandatory)
        :type: str
        :param clientOid: Unique order id created by users to identify their orders, e.g. UUID. (Mandatory)
        :type: str
        :param kwargs:  Fill in parameters with reference documents
        :return: {
                  "orderId": "5bd6e9286d99522a52e458de",
                  "borrowSize":10.2,
                  "loanApplyId":"600656d9a33ac90009de4f6f"
                }
        """
        params = {
            'symbol': symbol,
            'size': size,
            'side': side,
            'price': price,
            'type': "limit"
        }
        if not clientOid:
            clientOid = self.return_unique_id
        params['clientOid'] = clientOid
        if kwargs:
            params.update(kwargs)

        return self._request('POST', '/api/v1/margin/order', params=params)

    def create_market_margin_order(self, symbol, side, clientOid='', **kwargs):
        """
        https://docs.kucoin.com/#place-a-margin-order
        :param symbol: a valid trading symbol code (Mandatory)
        :type: str
        :param side: place direction buy or sell (Mandatory)
        :type: str
        :param clientOid: Unique order id created by users to identify their orders, e.g. UUID. (Mandatory)
        :type: str
        :param kwargs:  Fill in parameters with reference documents
        :return: {
                  "orderId": "5bd6e9286d99522a52e458de",
                  "borrowSize":10.2,
                  "loanApplyId":"600656d9a33ac90009de4f6f"
                }
        """
        params = {
            'symbol': symbol,
            'side': side,
            'type': "market"
        }
        if not clientOid:
            clientOid = self.return_unique_id
        params['clientOid'] = clientOid
        if kwargs:
            params.update(kwargs)

        return self._request('POST', '/api/v1/margin/order', params=params)

    def create_limit_order(self, symbol, side, size, price, clientOid='', **kwargs):
        """
        https://docs.kucoin.com/#place-a-new-order
        :param symbol: a valid trading symbol code (Mandatory)
        :type: str
        :param side: place direction buy or sell (Mandatory)
        :type: str
        :param size: amount of base currency to buy or sell (Mandatory)
        :type: str
        :param price: price per base currency (Mandatory)
        :type: str
        :param clientOid: Unique order id created by users to identify their orders, e.g. UUID. (Mandatory)
        :type: str
        :param kwargs:  Fill in parameters with reference documents
        :return: {'orderId': '5d9ee461f24b80689797fd04'}
        """
        params = {
            'symbol': symbol,
            'size': size,
            'side': side,
            'price': price,
            'type': "limit"
        }
        if not clientOid:
            clientOid = self.return_unique_id
        params['clientOid'] = clientOid
        if kwargs:
            params.update(kwargs)

        return self._request('POST', '/api/v1/orders', params=params)

    def create_limit_stop_order(self, symbol, side, size, price, stopPrice,  clientOid="", **kwargs):
        params = {
            'symbol': symbol,
            'size': size,
            'side': side,
            'price': price,
            'stopPrice': stopPrice,
            'type': "limit"
        }
        if not clientOid:
            clientOid = self.return_unique_id
        params['clientOid'] = clientOid
        if kwargs:
            params.update(kwargs)

        return self._request('POST', '/api/v1/stop-order', params=params)

    def create_market_stop_order(self, symbol, side, stopPrice, size="", funds="", clientOid="", **kwargs):
        params = {
            'symbol': symbol,
            'side': side,
            'stopPrice': stopPrice,
            'type': "market"

        }
        if not clientOid:
            clientOid = self.return_unique_id
        params['clientOid'] = clientOid
        if size:
            params['size'] = size
        elif funds:
            params['funds'] = funds
        else:
            raise Exception('Funds or size')

        if kwargs:
            params.update(kwargs)

        return self._request('POST', '/api/v1/stop-order', params=params)

    def create_market_order(self, symbol, side, clientOid='', **kwargs):
        """
        https://docs.kucoin.com/#place-a-new-order
        :param symbol: a valid trading symbol code (Mandatory)
        :type: str
        :param side: place direction buy or sell (Mandatory)
        :type: str
        :param clientOid: Unique order id created by users to identify their orders, e.g. UUID. (Mandatory)
        :type: str
        :param kwargs:  Fill in parameters with reference documents
        :return: {'orderId': '5d9ee461f24b80689797fd04'}
        """
        params = {
            'symbol': symbol,
            'side': side,
            'type': "market"
        }
        if not clientOid:
            clientOid = self.return_unique_id
        params['clientOid'] = clientOid
        if kwargs:
            params.update(kwargs)

        return self._request('POST', '/api/v1/orders', params=params)

    def create_bulk_orders(self, symbol, orderList):
        """
        https://docs.kucoin.com/#place-bulk-orders
        :param symbol: a valid trading symbol code.
        :type: str
        :param orderList: order list
        :type: list
        :return:
        {
          "data": [
            {
              "symbol": "KCS-USDT",
              "type": "limit",
              "side": "buy",
              "price": "0.01",
              "size": "0.01",
              "funds": null,
              "stp": "",
              "stop": "",
              "stopPrice": null,
              "timeInForce": "GTC",
              "cancelAfter": 0,
              "postOnly": false,
              "hidden": false,
              "iceberge": false,
              "iceberg": false,
              "visibleSize": null,
              "channel": "API",
              "id": "611a6a309281bc000674d3c0",
              "status": "success",
              "failMsg": null,
              "clientOid": "552a8a0b7cb04354be8266f0e202e7e9"
            },
            {
              "symbol": "KCS-USDT",
              "type": "limit",
              "side": "buy",
              "price": "0.01",
              "size": "0.01",
              "funds": null,
              "stp": "",
              "stop": "",
              "stopPrice": null,
              "timeInForce": "GTC",
              "cancelAfter": 0,
              "postOnly": false,
              "hidden": false,
              "iceberge": false,
              "iceberg": false,
              "visibleSize": null,
              "channel": "API",
              "id": "611a6a309281bc000674d3c1",
              "status": "success",
              "failMsg": null,
              "clientOid": "bd1e95e705724f33b508ed270888a4a9"
            }
          ]
        }
        """
        params = {
            'symbol': symbol,
            'orderList': orderList,
        }
        return self._request('POST', '/api/v1/orders/multi', params=params)

    def cancel_client_order(self, clientId):
        """
        :param orderId: str  (Mandatory)
        :return:{"cancelledOrderId": "5f311183c9b6d539dc614db3","clientOid": "6d539dc614db3"}
        """
        return self._request('DELETE', f'/api/v1/order/client-order/{clientId}')

    def cancel_stop_order(self, orderId):
        """
        :param orderId: Order ID, unique ID of the order. (Mandatory)
        :type: str
        :return:
        {
             "cancelledOrderIds": [
              "5bd6e9286d99522a52e458de"   //orderId
            ]
        }
        """
        return self._request('DELETE', f'/api/v1/stop-order/{orderId}')

    def cancel_client_stop_order(self, clientOid, symbol=""):
        """
        :param orderId: Order ID, unique ID of the order. (Mandatory)
        :type: str
        :return:
        {
             "cancelledOrderIds": [
              "5bd6e9286d99522a52e458de"   //orderId
            ]
        }
        """
        params = {
            "clientOid": clientOid
        }
        if symbol:
            params["symbol"] = symbol

        return self._request('DELETE', f'/api/v1/stop-order/cancelOrderByClientOid', params=params)

    def cancel_stop_condition_order(self, symbol="", tradeType="", orderIds=""):
        """
        """
        params = {}
        if symbol:
            params["symbol"] = symbol
        if tradeType:
            params["tradeType"] = tradeType
        if orderIds:
            params["orderIds"] = orderIds

        return self._request('DELETE', f'/api/v1/stop-order/cancel',params=params)

    def cancel_order(self, orderId):
        """
        https://docs.kucoin.com/#cancel-an-order

        :param orderId: Order ID, unique ID of the order. (Mandatory)
        :type: str
        :return:
        {
             "cancelledOrderIds": [
              "5bd6e9286d99522a52e458de"   //orderId
            ]
        }
        """
        return self._request('DELETE', '/api/v1/orders/{orderId}'.format(orderId=orderId))

    def cancel_all_orders(self, **kwargs):
        """
        https://docs.kucoin.com/#cancel-all-orders
        :param kwargs: [optional] symbol, tradeType
        :return:
        {
           "cancelledOrderIds": [
              "5c52e11203aa677f33e493fb",  //orderId
              "5c52e12103aa677f33e493fe",
              "5c52e12a03aa677f33e49401",
              "5c52e1be03aa677f33e49404",
              "5c52e21003aa677f33e49407",
              "5c6243cb03aa67580f20bf2f",
              "5c62443703aa67580f20bf32",
              "5c6265c503aa676fee84129c",
              "5c6269e503aa676fee84129f",
              "5c626b0803aa676fee8412a2"
            ]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('DELETE', '/api/v1/orders', params=params)

    def get_order_list(self, **kwargs):
        """
        https://docs.kucoin.com/#list-orders
        :param kwargs: [optional] symbol, status, side, type, tradeType, startAt, endAt, currentPage, pageSize and so on
        :return:
        {
            "currentPage": 1,
            "pageSize": 1,
            "totalNum": 153408,
            "totalPage": 153408,
            "items": [
              {
                "id": "5c35c02703aa673ceec2a168",   //orderid
                "symbol": "BTC-USDT",   //symbol
                "opType": "DEAL",      // operation type: DEAL
                "type": "limit",       // order type,e.g. limit,market,stop_limit.
                "side": "buy",         // transaction direction,include buy and sell
                "price": "10",         // order price
                "size": "2",           // order quantity
                "funds": "0",          // order funds
                "dealFunds": "0.166",  // deal funds
                "dealSize": "2",       // deal quantity
                "fee": "0",            // fee
                "feeCurrency": "USDT", // charge fee currency
                "stp": "",             // self trade prevention,include CN,CO,DC,CB
                "stop": "",            // stop type
                "stopTriggered": false,  // stop order is triggered
                "stopPrice": "0",      // stop price
                "timeInForce": "GTC",  // time InForce,include GTC,GTT,IOC,FOK
                "postOnly": false,     // postOnly
                "hidden": false,       // hidden order
                "iceberg": false,      // iceberg order
                "visibleSize": "0",    // display quantity for iceberg order
                "cancelAfter": 0,      // cancel orders time，requires timeInForce to be GTT
                "channel": "IOS",      // order source
                "clientOid": "",       // user-entered order unique mark
                "remark": "",          // remark
                "tags": "",            // tag order source
                "isActive": false,     // status before unfilled or uncancelled
                "cancelExist": false,   // order cancellation transaction record
                "createdAt": 1547026471000,  // create time
                "tradeType": "TRADE"
              }
            ]
         }
        """
        params = {}
        if kwargs:
            params.update(kwargs)

        return self._request('GET', '/api/v1/orders', params=params)

    def get_recent_orders(self):
        """
        https://docs.kucoin.com/#recent-orders
        :return:
        {
            "currentPage": 1,
            "pageSize": 1,
            "totalNum": 153408,
            "totalPage": 153408,
            "items": [
              {
                "id": "5c35c02703aa673ceec2a168",
                "symbol": "BTC-USDT",
                "opType": "DEAL",
                "type": "limit",
                "side": "buy",
                "price": "10",
                "size": "2",
                "funds": "0",
                "dealFunds": "0.166",
                "dealSize": "2",
                "fee": "0",
                "feeCurrency": "USDT",
                "stp": "",
                "stop": "",
                "stopTriggered": false,
                "stopPrice": "0",
                "timeInForce": "GTC",
                "postOnly": false,
                "hidden": false,
                "iceberg": false,
                "visibleSize": "0",
                "cancelAfter": 0,
                "channel": "IOS",
                "clientOid": "",
                "remark": "",
                "tags": "",
                "isActive": false,
                "cancelExist": false,
                "createdAt": 1547026471000,
                "tradeType": "TRADE"
              }
            ]
        }
        """

        return self._request('GET', '/api/v1/limit/orders')

    def get_order_details(self, orderId):
        """
        https://docs.kucoin.com/#get-an-order
        :param orderId: Order ID, unique identifier of an order, obtained via the List orders. (Mandatory)
        :return:
        {
            "id": "5c35c02703aa673ceec2a168",
            "symbol": "BTC-USDT",
            "opType": "DEAL",
            "type": "limit",
            "side": "buy",
            "price": "10",
            "size": "2",
            "funds": "0",
            "dealFunds": "0.166",
            "dealSize": "2",
            "fee": "0",
            "feeCurrency": "USDT",
            "stp": "",
            "stop": "",
            "stopTriggered": false,
            "stopPrice": "0",
            "timeInForce": "GTC",
            "postOnly": false,
            "hidden": false,
            "iceberg": false,
            "visibleSize": "0",
            "cancelAfter": 0,
            "channel": "IOS",
            "clientOid": "",
            "remark": "",
            "tags": "",
            "isActive": false,
            "cancelExist": false,
            "createdAt": 1547026471000,
            "tradeType": "TRADE"
        }
        """
        return self._request('GET', '/api/v1/orders/{orderId}'.format(orderId=orderId))

    def get_all_stop_order_details(self, **kwargs):
        """
        :param orderId: Order ID, unique identifier of an order, obtained via the List orders. (Mandatory)
        :return:
        {
            "id": "5c35c02703aa673ceec2a168",
            "symbol": "BTC-USDT",
            "opType": "DEAL",
            "type": "limit",
            "side": "buy",
            "price": "10",
            "size": "2",
            "funds": "0",
            "dealFunds": "0.166",
            "dealSize": "2",
            "fee": "0",
            "feeCurrency": "USDT",
            "stp": "",
            "stop": "",
            "stopTriggered": false,
            "stopPrice": "0",
            "timeInForce": "GTC",
            "postOnly": false,
            "hidden": false,
            "iceberg": false,
            "visibleSize": "0",
            "cancelAfter": 0,
            "channel": "IOS",
            "clientOid": "",
            "remark": "",
            "tags": "",
            "isActive": false,
            "cancelExist": false,
            "createdAt": 1547026471000,
            "tradeType": "TRADE"
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', f'/api/v1/stop-order', params=params)

    def get_stop_order_details(self, orderId):
        """
        :param orderId: Order ID, unique identifier of an order, obtained via the List orders. (Mandatory)
        :return:
        {
            "id": "5c35c02703aa673ceec2a168",
            "symbol": "BTC-USDT",
            "opType": "DEAL",
            "type": "limit",
            "side": "buy",
            "price": "10",
            "size": "2",
            "funds": "0",
            "dealFunds": "0.166",
            "dealSize": "2",
            "fee": "0",
            "feeCurrency": "USDT",
            "stp": "",
            "stop": "",
            "stopTriggered": false,
            "stopPrice": "0",
            "timeInForce": "GTC",
            "postOnly": false,
            "hidden": false,
            "iceberg": false,
            "visibleSize": "0",
            "cancelAfter": 0,
            "channel": "IOS",
            "clientOid": "",
            "remark": "",
            "tags": "",
            "isActive": false,
            "cancelExist": false,
            "createdAt": 1547026471000,
            "tradeType": "TRADE"
        }
        """
        return self._request('GET', f'/api/v1/stop-order/{orderId}')

    def get_client_stop_order_details(self, clientOid, symbol=''):
        """
        :param orderId: Order ID, unique identifier of an order, obtained via the List orders. (Mandatory)
        :return:
        {
            "id": "5c35c02703aa673ceec2a168",
            "symbol": "BTC-USDT",
            "opType": "DEAL",
            "type": "limit",
            "side": "buy",
            "price": "10",
            "size": "2",
            "funds": "0",
            "dealFunds": "0.166",
            "dealSize": "2",
            "fee": "0",
            "feeCurrency": "USDT",
            "stp": "",
            "stop": "",
            "stopTriggered": false,
            "stopPrice": "0",
            "timeInForce": "GTC",
            "postOnly": false,
            "hidden": false,
            "iceberg": false,
            "visibleSize": "0",
            "cancelAfter": 0,
            "channel": "IOS",
            "clientOid": "",
            "remark": "",
            "tags": "",
            "isActive": false,
            "cancelExist": false,
            "createdAt": 1547026471000,
            "tradeType": "TRADE"
        }
        """
        params = {
            "clientOid": clientOid
        }
        if symbol:
            params["symbol"] = symbol

        return self._request('GET', f'/api/v1/stop-order/queryOrderByClientOid', params=params)

    def get_fill_list(self, tradeType, **kwargs):
        """
        https://docs.kucoin.com/#list-fills
        :param tradeType: The type of trading (Mandatory)
        :param kwargs: [Optional] orderId, symbol, side, type, startAt, endAt, currentPage, pageSize
        :return:
        {
            "currentPage":1,
            "pageSize":1,
            "totalNum":251915,
            "totalPage":251915,
            "items":[
                {
                    "symbol":"BTC-USDT",    //symbol
                    "tradeId":"5c35c02709e4f67d5266954e",   //trade id
                    "orderId":"5c35c02703aa673ceec2a168",   //order id
                    "counterOrderId":"5c1ab46003aa676e487fa8e3",  //counter order id
                    "side":"buy",   //transaction direction,include buy and sell
                    "liquidity":"taker",  //include taker and maker
                    "forceTaker":true,  //forced to become taker
                    "price":"0.083",   //order price
                    "size":"0.8424304",  //order quantity
                    "funds":"0.0699217232",  //order funds
                    "fee":"0",  //fee
                    "feeRate":"0",  //fee rate
                    "feeCurrency":"USDT",  // charge fee currency
                    "stop":"",        // stop type
                    "type":"limit",  // order type,e.g. limit,market,stop_limit.
                    "createdAt":1547026472000,  //time
                    "tradeType": "TRADE"
                }
            ]
        }
        """
        params = {
            'tradeType': tradeType
        }
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/fills', params=params)

    def get_recent_fills(self):
        """
        https://docs.kucoin.com/#recent-fills
        :return:
        [
            {
                "counterOrderId":"5db7ee769797cf0008e3beea",
                "createdAt":1572335233000,
                "fee":"0.946357371456",
                "feeCurrency":"USDT",
                "feeRate":"0.001",
                "forceTaker":true,
                "funds":"946.357371456",
                "liquidity":"taker",
                "orderId":"5db7ee805d53620008dce1ba",
                "price":"9466.8",
                "side":"buy",
                "size":"0.09996592",
                "stop":"",
                "symbol":"BTC-USDT",
                "tradeId":"5db7ee8054c05c0008069e21",
                "tradeType":"MARGIN_TRADE",
                "type":"market"
            },
            {
                "counterOrderId":"5db7ee4b5d53620008dcde8e",
                "createdAt":1572335207000,
                "fee":"0.94625",
                "feeCurrency":"USDT",
                "feeRate":"0.001",
                "forceTaker":true,
                "funds":"946.25",
                "liquidity":"taker",
                "orderId":"5db7ee675d53620008dce01e",
                "price":"9462.5",
                "side":"sell",
                "size":"0.1",
                "stop":"",
                "symbol":"BTC-USDT",
                "tradeId":"5db7ee6754c05c0008069e03",
                "tradeType":"MARGIN_TRADE",
                "type":"market"
            },
            {
                "counterOrderId":"5db69aa4688933000aab8114",
                "createdAt":1572248229000,
                "fee":"1.882148318525",
                "feeCurrency":"USDT",
                "feeRate":"0.001",
                "forceTaker":false,
                "funds":"1882.148318525",
                "liquidity":"maker",
                "orderId":"5db69a9c4e6d020008f03275",
                "price":"9354.5",
                "side":"sell",
                "size":"0.20120245",
                "stop":"",
                "symbol":"BTC-USDT",
                "tradeId":"5db69aa477d8de0008c1efac",
                "tradeType":"MARGIN_TRADE",
                "type":"limit"
            }
        ]
        """
        return self._request('GET', '/api/v1/limit/fills')

    def get_client_order_details(self, clientOid):
        """
        https://docs.kucoin.com/#recent-fills
        :param clientOid: Unique order id created by users to identify their orders
        :return:
        {
          "id": "61149d589281bc00064a9ee0",
          "symbol": "KCS-USDT",
          "opType": "DEAL",
          "type": "limit",
          "side": "buy",
          "price": "0.001",
          "size": "0.01",
          "funds": "0",
          "dealFunds": "0",
          "dealSize": "0",
          "fee": "0",
          "feeCurrency": "USDT",
          "stp": "",
          "stop": "",
          "stopTriggered": false,
          "stopPrice": "0",
          "timeInForce": "GTC",
          "postOnly": false,
          "hidden": false,
          "iceberg": false,
          "visibleSize": "0",
          "cancelAfter": 0,
          "channel": "API",
          "clientOid": "03cd879961b64429b0e0149f311ce59f",
          "remark": null,
          "tags": null,
          "isActive": false,
          "cancelExist": true,
          "createdAt": 1628740952556,
          "tradeType": "MARGIN_TRADE"
        }
        """
        return self._request('GET', f'/api/v1/order/client-order/{clientOid}')


import shutil

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

try:
    shutil.rmtree('./build')
    shutil.rmtree('./dist')
except OSError as e:
    print("Error: %s - %s." % (e.filename, e.strerror))

setuptools.setup(
    name="TradeGate-Beta",  # This is the name of the package
    version="0.2.1Beta04",
    author="Rustin Soraki",  # Full name of the author
    description="A Trading Gateway",
    long_description=long_description,  # Long description read from the readme file
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(where='TradeGates'),  # List of all python modules to be installed
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],  # Information to filter the project on PyPi website
    python_requires='>=3.6',  # Minimum version requirement of the package
    py_modules=["TradeGate"],  # Name of the python package
    package_dir={'': 'TradeGates'},  # Directory of the source code of the package
    install_requires=['binance-connector', 'requests', 'apscheduler', 'websocket-client', 'urllib3', 'tzlocal<3.0',
                      'pandas', 'pybit']  # Install other dependencies if any
)

# python setup.py sdist bdist_wheel  
# twine check dist/*   
# twine upload dist/*



class Candlestick:

    def __init__(self):
        self.openTime = 0
        self.open = 0.0
        self.high = 0.0
        self.low = 0.0
        self.close = 0.0
        self.volume = 0.0
        self.closeTime = 0
        self.quoteAssetVolume = 0.0
        self.numTrades = 0
        self.takerBuyBaseAssetVolume = 0.0
        self.takerBuyQuoteAssetVolume = 0.0
        self.ignore = 0.0

    def toArray(self):
        outArr = []

        outArr.append(self.openTime)
        outArr.append(self.open)
        outArr.append(self.high)
        outArr.append(self.low)
        outArr.append(self.close)
        outArr.append(self.volume)
        outArr.append(self.closeTime)
        outArr.append(self.quoteAssetVolume)
        outArr.append(self.numTrades)
        outArr.append(self.takerBuyBaseAssetVolume)
        outArr.append(self.takerBuyQuoteAssetVolume)
        outArr.append(self.ignore)

        return outArr

    @staticmethod
    def json_parse(json_data):
        result = Candlestick()
        val = json_data.convert_2_list()
        result.openTime = val[0]
        result.open = float(val[1])
        result.high = float(val[2])
        result.low = float(val[3])
        result.close = float(val[4])
        result.volume = float(val[5])
        result.closeTime = val[6]
        result.quoteAssetVolume = float(val[7])
        result.numTrades = float(val[8])
        result.takerBuyBaseAssetVolume = float(val[9])
        result.takerBuyQuoteAssetVolume = float(val[10])
        result.ignore = val[11]
  
        return result


class FundingRate:

    def __init__(self):
        self.symbol = ""
        self.fundingRate = 0.0
        self.fundingTime = 0
    
    @staticmethod
    def json_parse(json_data):
        result = FundingRate()
        result.symbol = json_data.get_string("symbol")
        result.fundingRate = json_data.get_float("fundingRate")
        result.fundingTime = json_data.get_int("fundingTime")

        return result



class BLVTNAVCandlestick:

    def __init__(self):
        self.openTime = 0
        self.open = 0.0
        self.high = 0.0
        self.low = 0.0
        self.close = 0.0
        self.lev = 0.0  # Real leverage
        self.closeTime = 0
        self.numNAV = 0  # Number of NAV update

    @staticmethod
    def json_parse(json_data):
        result = BLVTNAVCandlestick()
        val = json_data.convert_2_list()
        result.openTime = val[0]
        result.open = val[1]
        result.high = val[2]
        result.low = val[3]
        result.close = val[4]
        result.lev = val[5]
        result.closeTime = val[6]
        result.numNAV = val[8]

        return result


import logging
import warnings
from datetime import datetime

from pybit import HTTP

from Exchanges.BaseExchange import BaseExchange
from Utils import DataHelpers, BybitHelpers


def isOrderDataValid(orderData: DataHelpers.OrderData):
    if orderData.symbol is None or orderData.quantity is None or orderData.side is None \
            or orderData.orderType is None:
        raise ValueError('Missing mandatory fields.')

    if orderData.orderType not in BybitExchange.spotOrderTypes:
        raise ValueError('Order type not correctly specified. Available order types for spot market: {}'.format(
            BybitExchange.spotOrderTypes))

    if orderData.side not in ['BUY', 'SELL']:
        raise ValueError('Order side can only be \'BUY\' or \'SELL\'')

    if orderData.timeInForce not in BybitExchange.spotTimeInForces:
        raise ValueError(
            'Time-in-force not correctly specified. Available time-in-force for spot market: {}'.format(
                BybitExchange.spotTimeInForces))

    if orderData.orderType in ['LIMIT', 'LIMIT_MAKER'] and orderData.price is None:
        raise ValueError('Price must be specified for limit orders.')


def isFuturesOrderDataValid(orderData: DataHelpers.futuresOrderData):
    if orderData.symbol is None:
        raise ValueError('Specify symbol.')
    if orderData.quantity is None:
        raise ValueError('Specify quantity.')
    if orderData.side is None:
        raise ValueError('Specify order side.')
    if orderData.orderType is None:
        raise ValueError('Specify order type.')
    if orderData.timeInForce is None:
        raise ValueError('Specify timeInForce.')
    if orderData.closePosition is None:
        raise ValueError('Specify closePosition.')
    if orderData.reduceOnly is None:
        raise ValueError('Specify reduceOnly.')

    if orderData.orderType not in BybitExchange.futuresOrderTypes:
        raise ValueError('Bad order type specified. Available order types for futures: {}'.format(
            BybitExchange.futuresOrderTypes))

    if orderData.orderType.startswith('STOP'):
        if orderData.extraParams is None:
            raise ValueError('Specify \'basePrice\' in \'extraParams\'')
        if 'basePrice' not in orderData.extraParams.keys():
            raise ValueError('Specify \'basePrice\' in \'extraParams\'')
        if orderData.stopPrice is None:
            raise ValueError('Specify \'stopPrice\'.')

    if 'LIMIT' in orderData.orderType and orderData.price is None:
        raise ValueError('Specify \'price\' for limit orders.')

    if orderData.timeInForce not in BybitExchange.futuresTimeInForces.keys() and \
            orderData.timeInForce not in BybitExchange.futuresTimeInForces.values():
        raise ValueError('\'timeInForce\' is not correct.')

    if orderData.extraParams is not None:
        if 'triggerBy' in orderData.extraParams.keys() \
                and orderData.extraParams['tpTriggerBy'] not in ['LastPrice', 'IndexPrice', 'MarkPrice']:
            raise ValueError('\'triggerBy\' was not correctly specified.')

        if 'tpTriggerBy' in orderData.extraParams.keys() \
                and orderData.extraParams['tpTriggerBy'] not in ['LastPrice', 'IndexPrice', 'MarkPrice']:
            raise ValueError('\'tpTriggerBy\' was not correctly specified.')

        if 'slTriggerBy' in orderData.extraParams.keys() \
                and orderData.extraParams['slTriggerBy'] not in ['LastPrice', 'IndexPrice', 'MarkPrice']:
            raise ValueError('\'slTriggerBy\' was not correctly specified.')

        if 'positionIdx' in orderData.extraParams.keys() and orderData.extraParams['positionIdx'] not in [0, 1, 2]:
            raise ValueError('\'positionIdx\' was not correctly specified.')


class PyBitHTTP(HTTP):
    def __init__(self, endpoint=None, api_key=None, api_secret=None, logging_level=logging.INFO, log_requests=False,
                 request_timeout=10, recv_window=5000, force_retry=False, retry_codes=None, ignore_codes=None,
                 max_retries=3, retry_delay=3, referral_id=None, spot=False):
        super().__init__(endpoint, api_key, api_secret, logging_level, log_requests, request_timeout, recv_window,
                         force_retry, retry_codes, ignore_codes, max_retries, retry_delay, referral_id, spot)

    def query_history_order(self, **kwargs):
        if self.spot is True:
            suffix = '/spot/v1/history-orders'

            return self._submit_request(
                method='GET',
                path=self.endpoint + suffix,
                query=kwargs,
                auth=True
            )
        else:
            raise NotImplementedError('Not implemented for futures market.')

    def get_active_order_spot(self, **kwargs):
        if self.spot is True:
            suffix = '/spot/v1/order'

            return self._submit_request(
                method='GET',
                path=self.endpoint + suffix,
                query=kwargs,
                auth=True
            )
        else:
            raise NotImplementedError('Not implemented for futures market.')


class BybitExchange(BaseExchange):
    timeIndexesInCandleData = [0, 6]
    desiredCandleDataIndexes = [0, 1, 2, 3, 4, 5, 6, 8]

    spotOrderTypes = ['LIMIT', 'MARKET', 'LIMIT_MAKER']
    spotTimeInForces = ['GTC', 'FOK', 'IOC']

    futuresOrderTypes = ['MARKET', 'LIMIT', 'STOP_MARKET', 'STOP_LIMIT']
    futuresTimeInForces = {'GTC': 'GoodTillCancel', 'IOC': 'ImmediateOrCancel', 'FIK': 'FillOrKill', 'PO': 'PostOnly'}

    timeIntervals = ['1m', '3m', '5m', '15m', '30m', '1h', '2h', '4h', '6h', '12h', '1d', '1w', '1M']

    def __init__(self, credentials, sandbox=False, unifiedInOuts=True):
        self.apiKey = credentials['spot']['key']
        self.secret = credentials['spot']['secret']
        self.sandbox = sandbox
        self.unifiedInOuts = unifiedInOuts

        if sandbox:
            self.spotSession = PyBitHTTP("https://api-testnet.bybit.com", api_key=self.apiKey, api_secret=self.secret,
                                         spot=True)
            self.futuresSession = PyBitHTTP("https://api-testnet.bybit.com", api_key=self.apiKey,
                                            api_secret=self.secret)
        else:
            self.spotSession = PyBitHTTP("https://api.bybit.com", api_key=self.apiKey, api_secret=self.secret,
                                         spot=True)
            self.futuresSession = PyBitHTTP("https://api.bybit.com", api_key=self.apiKey, api_secret=self.secret)

        self.futuresSymbols = []
        for symbol in self.futuresSession.query_symbol()['result']:
            if symbol['name'].endswith('USDT'):
                self.futuresSymbols.append(symbol['name'])

    def getBalance(self, asset='', futures=False):
        if futures:
            if asset in [None, '']:
                return BybitHelpers.getBalanceOut(self.futuresSession.get_wallet_balance()['result'], futures=True)
            else:
                return BybitHelpers.getBalanceOut(self.futuresSession.get_wallet_balance(coin=asset)['result'],
                                                  single=True, futures=True)
        else:
            if asset in [None, '']:
                return BybitHelpers.getBalanceOut(self.spotSession.get_wallet_balance()['result']['balances'])
            else:
                assets = self.spotSession.get_wallet_balance()['result']['balances']
                for coin in assets:
                    if asset == coin['coin']:
                        return BybitHelpers.getBalanceOut(coin, single=True)

                try:
                    self.futuresSession.get_wallet_balance(coin=asset)
                    return BybitHelpers.makeDummyBalance(asset)
                except Exception:
                    raise ValueError('Coin not found.')

    def symbolAccountTradeHistory(self, symbol, futures=False, fromId=None, limit=None):
        if futures:
            tradeHistory = self.futuresSession.user_trade_records(symbol=symbol, limit=limit, fromId=fromId)
            return BybitHelpers.getMyTradeHistoryOut(tradeHistory['result']['data'], futures=True)
        else:
            tradeHistory = self.spotSession.user_trade_records(symbol=symbol, limit=limit, fromId=fromId)
            return BybitHelpers.getMyTradeHistoryOut(tradeHistory['result'])

    def testSpotOrder(self, orderData: DataHelpers.OrderData):
        isOrderDataValid(orderData)

        if orderData.icebergQty is not None or orderData.newOrderRespType is not None \
                or orderData.quoteOrderQty is not None or orderData.recvWindow is not None \
                or orderData.stopPrice is not None:
            warnings.warn('Some of the given parameters have no use in ByBit exchange.')

        return orderData

    def makeSpotOrder(self, orderData):
        orderParams = BybitHelpers.getSpotOrderAsDict(orderData)

        return BybitHelpers.getMakeSpotOrderOut(self.spotSession.place_active_order(**orderParams)['result'])

    def getSymbolOrders(self, symbol, futures=False, orderId=None, startTime=None, endTime=None, limit=None):
        if futures:
            historyList = []
            pageNumber = 1
            endTimeString = None
            startTimeString = None
            done = False
            while not done:
                history = self.futuresSession.get_active_order(symbol=symbol, page=pageNumber, limit=50)

                if startTime is not None:
                    startTimeString = startTime.strftime('%Y-%m-%dT%H:%M:%SZ')
                if endTime is not None:
                    endTimeString = endTime.strftime('%Y-%m-%dT%H:%M:%SZ')

                for order in history['result']['data']:
                    if endTime is not None:
                        if endTimeString < order['create_time']:
                            continue

                    if startTime is not None:
                        if order['created_time'] < startTimeString:
                            done = True
                            break

                    historyList.append(order)

                if limit is not None and limit <= len(historyList):
                    done = True

                if len(history['result']['data']) < 50:
                    done = True

                pageNumber += 1

            return historyList
        else:
            history = self.spotSession.query_history_order(symbol=symbol, orderId=orderId, startTime=startTime,
                                                           endtime=endTime, limit=limit)
            return history['result']

    def getOpenOrders(self, symbol, futures=False):
        if futures:
            openOrders = []

            openActiveOrders = self.futuresSession.query_active_order(symbol=symbol)
            for activeOrder in openActiveOrders['result']:
                openOrders.append(BybitHelpers.futuresOrderOut(activeOrder))

            openConditionalOrders = self.futuresSession.query_conditional_order(symbol=symbol)
            for conditionalOrder in openConditionalOrders['result']:
                openOrders.append(BybitHelpers.futuresOrderOut(conditionalOrder, isConditional=True))

            return openOrders
        else:
            if symbol is None:
                openOrders = self.spotSession.query_active_order()['result']
            else:
                openOrders = self.spotSession.query_active_order(symbol=symbol)['result']
            return BybitHelpers.getOpenOrdersOut(openOrders)

    def cancelAllSymbolOpenOrders(self, symbol, futures=False):
        if futures:
            canceledOrdersIds = []
            result = self.futuresSession.cancel_all_active_orders(symbol=symbol)
            canceledOrdersIds.append(result['result'])

            result = self.futuresSession.cancel_all_conditional_orders(symbol=symbol)
            canceledOrdersIds.append(result['result'])
        else:
            result = self.spotSession.batch_fast_cancel_active_order(symbol=symbol,
                                                                     orderTypes="LIMIT,LIMIT_MAKER,MARKET")
            return result['result']['success']

    def cancelOrder(self, symbol, orderId=None, localOrderId=None, futures=False):
        if futures:
            if orderId is not None:
                try:
                    result = self.futuresSession.cancel_active_order(symbol=symbol, order_id=orderId)
                except Exception as e:
                    try:
                        result = self.futuresSession.cancel_conditional_order(symbol=symbol, order_id=orderId)
                    except Exception as e:
                        raise RuntimeError('Problem in canceling order in bybit: {}'.format(str(e)))
            elif localOrderId is not None:
                try:
                    result = self.futuresSession.cancel_active_order(symbol=symbol, order_link_id=localOrderId)
                except Exception as e:
                    try:
                        result = self.futuresSession.cancel_conditional_order(symbol=symbol, order_link_id=localOrderId)
                    except Exception as e:
                        raise RuntimeError('Problem in canceling order in bybit: {}'.format(str(e)))
            else:
                raise ValueError('Must specify either \'orderId\' or \'localOrderId\'')

            return result['result']
        else:
            if orderId is not None:
                result = self.spotSession.cancel_active_order(orderId=orderId)
            elif localOrderId is not None:
                result = self.spotSession.cancel_active_order(orderLinkId=localOrderId)
            else:
                raise ValueError('Must specify either \'orderId\' or \'localOrderId\'')
            return result['result']

    def getOrder(self, symbol, orderId=None, localOrderId=None, futures=False):
        if futures:
            if orderId is not None:
                try:
                    order = self.futuresSession.query_active_order(symbol=symbol, order_id=orderId)
                except Exception as e:
                    try:
                        order = self.futuresSession.query_conditional_order(symbol=symbol, order_id=orderId)
                    except Exception as e:
                        raise RuntimeError('Problem in fetching order from bybit: {}'.format(str(e)))
            elif localOrderId is not None:
                try:
                    order = self.futuresSession.query_active_order(symbol=symbol, order_link_id=localOrderId)
                except Exception as e:
                    try:
                        order = self.futuresSession.query_conditional_order(symbol=symbol, order_link_id=localOrderId)
                    except Exception as e:
                        raise RuntimeError('Problem in fetching order from bybit: {}'.format(str(e)))
            else:
                raise ValueError('Specify either order Id in the exchange or local Id sent with the order')

            return BybitHelpers.futuresOrderOut(order['result'])
        else:
            if orderId is not None:
                try:
                    order = self.spotSession.get_active_order_spot(orderId=orderId)['result']
                except Exception as e:
                    raise RuntimeError('Problem in fetching order from bybit.')
            elif localOrderId is not None:
                try:
                    order = self.spotSession.get_active_order_spot(orderLinkId=localOrderId)['result']
                except Exception as e:
                    raise RuntimeError('Problem in fetching order from bybit.')
            else:
                raise ValueError('Specify either order Id in the exchange or local Id sent with the order')

            return BybitHelpers.getOrderOut(order)

    def getTradingFees(self):
        raise NotImplementedError()

    def getSymbolTickerPrice(self, symbol, futures=False):
        if futures:
            symbolInfo = self.futuresSession.latest_information_for_symbol(symbol=symbol)['result']
            return float(symbolInfo[0]['last_price'])
        else:
            symbolInfo = self.spotSession.latest_information_for_symbol(symbol=symbol)
            return float(symbolInfo['result']['lastPrice'])

    def getSymbolKlines(self, symbol, interval, startTime=None, endTime=None, limit=None, futures=False, blvtnav=False,
                        convertDateTime=False, doClean=False, toCleanDataframe=False):
        if interval not in self.timeIntervals:
            raise ValueError('Time interval is not valid.')

        if futures:
            data = self._getFuturesSymbolKlines(interval, limit, startTime, symbol)
        else:
            data = self._getSpotSymbolKlines(endTime, interval, limit, startTime, symbol)

        if convertDateTime or toCleanDataframe:
            BybitHelpers.klinesConvertDate(data, futures, self.timeIndexesInCandleData)

        if doClean or toCleanDataframe:
            finalDataArray = BybitHelpers.getKlinesDesiredOnlyCols(data, self.desiredCandleDataIndexes)

            if toCleanDataframe:
                return BybitHelpers.klinesConvertToPandas(finalDataArray)
            return finalDataArray
        else:
            return data

    def _getSpotSymbolKlines(self, endTime, interval, limit, startTime, symbol):
        if startTime is not None:
            startTimestamp = startTime.timestamp() * 1000
        else:
            startTimestamp = None
        if endTime is not None:
            endTimestamp = endTime.timestamp() * 1000
        else:
            endTimestamp = None

        if limit is not None:
            if limit > 1000:
                limit = 1000
            elif limit < 1:
                limit = 1
        data = self.spotSession.query_kline(symbol=symbol, interval=interval, startTime=startTimestamp,
                                            endTime=endTimestamp, limit=limit)['result']
        for datum in data:
            for idx in range(len(datum)):
                if idx in self.timeIndexesInCandleData:
                    continue
                datum[idx] = float(datum[idx])
        return data

    def _getFuturesSymbolKlines(self, interval, limit, startTime, symbol):
        futuresInterval = BybitHelpers.convertIntervalToFuturesKlines(interval)
        data = []
        if limit is not None:
            if limit > 200:
                limit = 200
            elif limit < 1:
                limit = 1
        else:
            limit = 200
        if startTime is None:
            startTimestamp = int(
                datetime.now().timestamp() - BybitHelpers.getIntervalInSeconds(interval, self.timeIntervals) * limit)
        else:
            startTimestamp = int(startTime.timestamp)
        candles = self.futuresSession.query_kline(symbol=symbol, interval=futuresInterval, from_time=startTimestamp,
                                                  limit=limit)
        for candle in candles['result']:
            dataArray = [float(candle['open_time']), float(candle['open']), float(candle['high']),
                         float(candle['low']), float(candle['close']), float(candle['volume']),
                         int(candle['open_time']) + BybitHelpers.getIntervalInSeconds(interval, self.timeIntervals),
                         None, None, None, None]
            data.append(dataArray)
        return data

    def getExchangeTime(self, futures=False):
        if futures:
            return self.futuresSession.server_time()['time_now']
        else:
            return int(self.spotSession.server_time()['result']['serverTime'])

    def getSymbol24hTicker(self, symbol):
        raise NotImplementedError()

    def testFuturesOrder(self, futuresOrderData):
        if futuresOrderData.timeInForce is None:
            futuresOrderData.timeInForce = 'GoodTillCancel'

        if futuresOrderData.closePosition is None:
            futuresOrderData.closePosition = False

        if futuresOrderData.reduceOnly is None:
            futuresOrderData.reduceOnly = False

        isFuturesOrderDataValid(futuresOrderData)

        return futuresOrderData

    def makeFuturesOrder(self, futuresOrderData: DataHelpers.futuresOrderData):
        orderParams = BybitHelpers.getFuturesOrderAsDict(futuresOrderData, self.futuresTimeInForces)

        if 'STOP' in futuresOrderData.orderType:
            result = self.futuresSession.place_conditional_order(**orderParams)
            return BybitHelpers.futuresOrderOut(result['result'], isConditional=True)
        else:
            result = self.futuresSession.place_active_order(**orderParams)
            return BybitHelpers.futuresOrderOut(result['result'])

    def makeBatchFuturesOrder(self, futuresOrderDatas):
        batchOrders = []
        batchConditionalOrders = []
        for order in futuresOrderDatas:
            orderAsDict = BybitHelpers.getFuturesOrderAsDict(order, self.futuresTimeInForces)

            if 'STOP' in order.orderType:
                batchConditionalOrders.append(order)
            else:
                batchOrders.append(orderAsDict)

        results = []
        if len(batchConditionalOrders) > 0:
            putResults = self.futuresSession.place_conditional_order_bulk(batchConditionalOrders)
            for result in putResults:
                results.append(BybitHelpers.futuresOrderOut(result['result'], isConditional=True))
        if len(batchOrders) > 0:
            putResults = self.futuresSession.place_active_order_bulk(batchOrders)
            for result in putResults:
                results.append(BybitHelpers.futuresOrderOut(result['result']))

        return results

    def changeInitialLeverage(self, symbol, leverage):
        return self.futuresSession.set_leverage(symbol=symbol, leverage=leverage)['result']

    def changeMarginType(self, symbol, marginType, params):
        try:
            buyLeverage = params['buyLeverage']
            sellLeverage = params['sellLeverage']
            if marginType.upper() == 'ISOLATED':
                isIsolated = True
            elif marginType.upper() == 'CROSS':
                isIsolated = False
            else:
                raise ValueError('Margin type must either be \'ISOLATED\' or \'CROSS\'.')
        except Exception:
            raise ValueError('Must specify \'buyLeverage\' and \'sellLeverage\' in \'params')

        self.futuresSession.cross_isolated_margin_switch(symbol=symbol, is_isolated=isIsolated,
                                                         buy_leverage=buyLeverage, sell_leverage=sellLeverage)
        return True

    def changePositionMargin(self, symbol, amount, marginType=None):
        return self.futuresSession.change_margin(symbol=symbol, margin=amount)['result']

    def getPosition(self):
        return self.futuresSession.my_position()['result']

    def spotBestBidAsks(self, symbol=None):
        return self.spotSession.best_bid_ask_price(symbol=symbol)['result']

    def getSymbolOrderBook(self, symbol, limit=None, futures=False):
        if futures:
            return self.futuresSession.orderbook(symbol=symbol)['result']
        else:
            return self.spotSession.orderbook(symbol=symbol)['result']

    def getSymbolRecentTrades(self, symbol, limit=None, futures=False):
        if futures:
            if limit is not None and limit > 0:
                limit = 1000 if limit > 1000 else limit
            else:
                limit = 500

            recentTrades = self.futuresSession.public_trading_records(symbol=symbol, limit=limit)['result']
            return BybitHelpers.getRecentTradeHistoryOut(recentTrades, futures=True)
        else:
            if limit is not None and limit > 0:
                limit = 60 if limit > 60 else limit
            else:
                limit = 60

            recentTrades = self.spotSession.public_trading_records(symbol=symbol, limit=limit)['result']
            return BybitHelpers.getRecentTradeHistoryOut(recentTrades)

    def getPositionInfo(self, symbol=None):
        result = self.futuresSession.my_position(symbol=symbol)
        return result['result']

    def getSymbolMinTrade(self, symbol, futures=False):
        symbolTickerPrice = self.getSymbolTickerPrice(symbol=symbol, futures=futures)

        minQuantity = None
        minQuoteQuantity = None
        stepQuantity = None
        stepPrice = None

        if futures:
            symbolInfos = self.futuresSession.query_symbol()['result']

            for symbolInfo in symbolInfos:
                if symbolInfo['name'] == symbol:
                    minQuantity = float(symbolInfo['lot_size_filter']['min_trading_qty'])
                    minQuoteQuantity = symbolTickerPrice * minQuantity
                    stepQuantity = float(symbolInfo['lot_size_filter']['qty_step'])
                    stepPrice = symbolInfo['price_filter']['tick_size']

        else:
            symbolInfos = self.spotSession.query_symbol()['result']

            for symbolInfo in symbolInfos:
                if symbolInfo['name'] == symbol:
                    minQuantity = float(symbolInfo['minTradeQuantity'])
                    minQuoteQuantity = float(symbolInfo['minTradeAmount'])
                    stepQuantity = float(symbolInfo['basePrecision'])
                    stepPrice = symbolInfo['minPricePrecision']

        return {'minQuantity': minQuantity, 'minQuoteQuantity': minQuoteQuantity,
                'precisionStep': stepQuantity, 'stepPrice': stepPrice}



class CandlestickInterval:
    MIN1 = "1m"
    MIN3 = "3m"
    MIN5 = "5m"
    MIN15 = "15m"
    MIN30 = "30m"
    HOUR1 = "1h"
    HOUR2 = "2h"
    HOUR4 = "4h"
    HOUR6 = "6h"
    HOUR8 = "8h"
    HOUR12 = "12h"
    DAY1 = "1d"
    DAY3 = "3d"
    WEEK1 = "1w"
    MON1 = "1m"
    INVALID = None


class OrderSide:
    BUY = "BUY"
    SELL = "SELL"
    INVALID = None


class TimeInForce:
    GTC = "GTC"
    IOC = "IOC"
    FOK = "FOK"
    GTX = "GTX"
    INVALID = None


class TradeDirection:
    BUY = "buy"
    SELL = "sell"
    INVALID = None


class OrderType:
    LIMIT = "LIMIT"
    MARKET = "MARKET"
    STOP = "STOP"
    STOP_MARKET = "STOP_MARKET"
    TAKE_PROFIT = "TAKE_PROFIT"
    TAKE_PROFIT_MARKET = "TAKE_PROFIT_MARKET"
    TRAILING_STOP_MARKET = "TRAILING_STOP_MARKET"
    INVALID = None

class OrderRespType:
    ACK = "ACK"
    RESULT = "RESULT"
    INVALID = None


class MatchRole:
    MAKER = "maker"
    TAKER = "taker"

class DepthStep:
    STEP0 = "step0"
    STEP1 = "step1"
    STEP2 = "step2"
    STEP3 = "step3"
    STEP4 = "step4"
    STEP5 = "step5"


class SubscribeMessageType:
    RESPONSE = "response"
    PAYLOAD = "payload"


class TransferType:
    ROLL_IN = "ROLL_IN"
    ROLL_OUT = "ROLL_OUT"
    INVALID = None

class WorkingType:
    MARK_PRICE = "MARK_PRICE"
    CONTRACT_PRICE = "CONTRACT_PRICE"
    INVALID = None


class FuturesMarginType:
    ISOLATED = "ISOLATED"
    CROSSED = "CROSSED"

class PositionSide:
    BOTH = "BOTH"
    LONG = "LONG"
    SHORT = "SHORT"
    INVALID = None



class IncomeType:
    TRANSFER = "TRANSFER"
    WELCOME_BONUS = "WELCOME_BONUS"
    REALIZED_PNL = "REALIZED_PNL"
    FUNDING_FEE = "FUNDING_FEE"
    COMMISSION = "COMMISSION"
    INSURANCE_CLEAR = "INSURANCE_CLEAR"
    INVALID = None

class UpdateTime:
    NORMAL = ""
    FAST = "@100ms"
    REALTIME = "@0ms"
    INVALID = ""

class ContractType:
    PERPETUAL = "perpetual"
    CURRENT_MONTH = "current_month"
    NEXT_MONTH = "next_month"


